from torchtext.data import Iterator
import random



class BucketIterator(Iterator):
    '''
    BucketIterator is torchtext Iterator which has capability to reduce padding and sorting batches in epochs
    '''
    def create_batches(self):
        self.batches = self.pool(self.data(), self.batch_size,
                                self.sort_key, self.batch_size_fn,
                                random_shuffler=self.random_shuffler,
                                shuffle=self.shuffle)

    def batch(self, data, batch_size, batch_size_fn=None):
        if batch_size_fn is None:
            batch_size_fn = lambda new, count, sofar: count
        minibatch, size_so_far = [], 0
        for ex in data:
            minibatch.append(ex)
            size_so_far = batch_size_fn(ex, len(minibatch), size_so_far)
            if size_so_far == batch_size:
                yield minibatch
                minibatch, size_so_far = [], 0
            elif size_so_far > batch_size:
                yield minibatch[:-1]
                minibatch, size_so_far = minibatch[-1:], batch_size_fn(ex, 1, 0)
        if minibatch:
            yield minibatch

    def pool(self, data, batch_size, key, batch_size_fn,
             random_shuffler,shuffle):
        p_batch = self.batch(sorted(data, key=key), batch_size, batch_size_fn)
        if shuffle:
            for b in random_shuffler(list(p_batch)):
                yield b
        else:
            for b in list(p_batch):
                yield b
                
    def __len__(self):
        p_batch = self.batch(sorted(self.data(), key=self.sort_key), self.batch_size, self.batch_size_fn)
        return len(list(p_batch))
    
def conf_bsf(batch_size,field_name):
    '''
    configurator of batch size function which help us to utilize all GPU card capability by creating batch based on number of tokens
    batch_size - how many tokens should appear with every batch
    field_name - field_name in torchtext Iterator base on which batch_size should be computed
    '''
    def bsf(new, count, sofar):
        '''
        new - batch example which will be added to batch
        count - the true size of batch
        sofar - the size on batch computing by this function after previous example

        return new batch size bese which is sum of exaples lengths (sum of tokens in batch)
        '''
        example_len = len(getattr(new,field_name))
        if count == 1 and example_len > batch_size:
            return batch_size
        if count == 1 or sofar/(count-1) == example_len:
            return example_len+sofar
        else:
            return example_len+sofar+(count-1)*(example_len-sofar/(count-1))
    return bsf


import time
import pdb
'''
TODO there should be developed class which can agreggate varius early stopping function in one instead of writing configurator per every stopping combination
'''


def conf_early_stopping_metric(patient,tolerance,max_patient):
    
    tolerance = 1-tolerance
    def early_stopping(obj):
        intervals_patient = int(obj.intervals_in_epoch * patient)
        max_intervals_patient = int(obj.intervals_in_epoch * max_patient)
        if intervals_patient >= len(obj.v_metrics[obj.names[0]]):
            return False, None

        best_metric = obj.metric[0].get_best(obj.v_metrics[obj.names[0]])
        best_pos = obj.v_metrics[obj.names[0]].index(best_metric)
        if len(obj.v_metrics[obj.names[0]]) - best_pos > intervals_patient:
            tolerant_best = best_metric * tolerance
            actual_metric = obj.v_metrics[obj.names[0]][-1]
            if obj.metric[0].get_best([tolerant_best,actual_metric]) == tolerant_best:
                return True, 'Lack of metric improvement.'
            if obj.metric[0].get_best([best_metric,actual_metric]) == best_metric and max_intervals_patient <= len(obj.v_metrics[obj.names[0]]):
                return True, 'Lack of metric improvement.'
        return False, None
    return early_stopping


def conf_early_stopping_loss(patient,tolerance,max_patient):
    tolerance = 1-tolerance
    def early_stopping(obj):
        # intervals_patient could be pre-computed in conf or somewhere else
        intervals_patient = int(obj.intervals_in_epoch * patient)
        max_intervals_patient = int(obj.intervals_in_epoch * max_patient)
        if intervals_patient >= len(obj.v_losses[obj.names[0]]):
            return False, None
        best_loss = min(obj.v_losses[obj.names[0]])
        best_pos = obj.v_losses[obj.names[0]].index(best_loss)
        if len(obj.v_losses[obj.names[0]]) - best_pos > intervals_patient:
            tolerant_best = best_loss * tolerance
            actual_loss = obj.v_losses[obj.names[0]][-1]
            if min([tolerant_best,actual_loss]) == tolerant_best:
                return True, 'Lack of loss decrease.'
            if min([best_loss,actual_loss]) == best_loss and max_intervals_patient <= len(obj.v_losses[obj.names[0]]):
                return True, 'Lack of loss decrease.'
        return False, None
    return early_stopping


def conf_early_stopping_all(patient,tolerance,max_patient):
    es_loss = conf_early_stopping_loss(patient,tolerance,max_patient)
    es_metric = conf_early_stopping_metric(patient,tolerance,max_patient)
    def early_stopping(obj):
        stop1, reason1 = es_loss(obj)
        stop2, reason2 = es_metric(obj)
        if stop1 and stop2:
            print('Stopping - lack of improvement.')
            return True, '{} {}'.format(reason1,reason2)
        else:
            return False
    return early_stopping

def conf_early_stopping_all_time(patient,tolerance,max_patient,time_stop):
    es_loss = conf_early_stopping_loss(patient,tolerance,max_patient)
    es_metric = conf_early_stopping_metric(patient,tolerance,max_patient)
    def early_stopping(obj):
        stop1, reason1 = es_loss(obj)
        stop2, reason2 = es_metric(obj)
        if stop1 and stop2:
            print('Stopping - lack of improvement.')
            return True, '{} {}'.format(reason1,reason2)
        elif time.time() - obj.time_start > time_stop:
            print('Stopping - too time consuming.')
            return True, 'Too time consuming.'
        else:
            return False, None
    return early_stopping


import numpy as np

class LR_Search:
    '''
    class which can check best learning rate for particular architecture 

    NEEDS TO BE REWRITED BC UNFORTUNATELY IT STRUGGLE TO FIND BEST LEARNING RATE
    '''
    
    def __init__(self, gym, check_interval, max_lr=2, min_lr=0.000001, max_lr_steps=0, min_lr_steps=0,valid_frac=1,diff_lr=None):
        self.gym = gym
        self.check_interval = check_interval
        self.valid_frac = valid_frac
        self.diff_lr = diff_lr
        if self.diff_lr:
            self.gym.optimizer_update = conf_update_with_diff(diff_lr)
        self.lr_bounds = self.fit_lr(max_lr, min_lr, max_lr_steps, min_lr_steps)

    def check_lr(self):
        i = 0
        lr_losses = []
        print("Searching learning rate...")
        last_loss = float('Inf')
        while True:
            lr = self.get_lr(self.min_lr, i)
            if lr > self.max_lr:
                break
            else:
                self.gym.optimizer.defaults['lr'] = lr
            print(lr)
            results = self.gym.workout(self.check_interval,stop_after_one=True,valid_frac=self.valid_frac,save_state=False)
            lr_loss = results['train_losses'][0]
            if np.isnan(lr_loss):
                break
            lr_losses.append([lr, lr_loss])
            i += 1
#             if last_loss * 1.1 < lr_loss:
#                 break
            last_loss = lr_loss
        return lr_losses

    def get_lr(self,min_lr,i):
        return min_lr * np.exp(i)

    def choose_min_lr(self, lr_losses):
        losses = np.array(lr_losses)[:, 1:2]
        ratios = []
        for i in range(len(losses) - 1):
            ratio = losses[i + 1] / losses[i]
            ratios.append(ratio[0])
        return ratios.index(min(ratios)) + 1

    def choose_max_lr(self, lr_losses, threshold=0.1):
        losses = np.array(lr_losses)[:, 1:2]
        rev_losses = losses[::-1]
        for i in range(len(rev_losses) - 1):
            ratio = rev_losses[i] / rev_losses[i + 1]
            if 1 - ratio > threshold:
                return losses.tolist().index(rev_losses[i])
        return len(lr_losses) - 1

    def fit_lr(self, max_lr, min_lr, max_lr_steps, min_lr_steps):
        self.max_lr = max_lr
        self.min_lr = min_lr
        self.max_lr_steps = max_lr_steps
        self.min_lr_steps = min_lr_steps

        lr_losses = self.check_lr()
        min_lr_idx = self.choose_min_lr(lr_losses) + self.min_lr_steps
        min_lr = lr_losses[min(max(0, min_lr_idx),len(lr_losses)-2)][0]
        max_lr_idx = self.choose_max_lr(lr_losses) - self.max_lr_steps
        max_lr_idx = min(max(min_lr_idx + 1, max_lr_idx),len(lr_losses)-1)
        max_lr = lr_losses[max_lr_idx][0]
        print("Chosen lr between " + min_lr.__str__() + ' and ' + max_lr.__str__())
        self.gym.optimizer.defaults['lr'] = min_lr
        return {'min_lr':min_lr,'max_lr':max_lr}

def conf_update_with_diff(diff_ratios):
    '''
    configuration of differential learning rate with ratios (optimizer needs to be created with parameters dict)
    diff_ratios - list of ratios per optimizers param_groups, ratio is ratio to deafault learning rate in optimizer
    '''
    def update_with_diff(obj):
        for p, lr_ratio in zip(obj.optimizer.param_groups,diff_ratios):
            p['lr'] = obj.optimizer.defaults['lr']*lr_ratio
    return update_with_diff




import numpy as np
import pandas as pd

from torch import nn

from tqdm import tqdm
tqdm.pandas()

from sklearn.neighbors import NearestNeighbors 

'''
ErrorAnalysis in beta version, strongly dependent on trochtext Iterator and can only work if dataset has only one field with text
'''

class ErrorAnalysis:
    def __init__(self, gym, join_pattern=' '):
        self.gym = gym
        batch_size = self.gym.valid.batch_size
        batch_size_fn = self.gym.valid.batch_size_fn
        self.gym.valid.batch_size = 1
        self.gym.valid.batch_size_fn = None
        if not hasattr(self.gym, 'model'):
            self.gym.model = self.gym.model_origin
        
        self.gym.model.to(gym.device)
        self.vactors = None
        valid_len = len(self.gym.valid)
        text_l = [None] * valid_len
        score_l = [None] * valid_len
        loss_l = [None] * valid_len
        preds_l = [None] * valid_len
        true_l = [None] * valid_len
        for i, data in enumerate(tqdm(self.gym.valid)):
            out,lab,loss = self.gym.data_flow(self.gym,data)
            loss_l[i] = loss.item()
            score_l[i] = self.gym.metric.evaluate([out],[lab])
            preds_l[i] = dict(zip(data.dataset.fields[data.target_fields[0]].vocab.itos,nn.functional.softmax(out,dim=-1).detach().cpu().numpy()[0]))
            true_l[i] = data.dataset.fields[data.target_fields[0]].vocab.itos[lab.item()]
            t,_=data
            new_text = join_pattern.join([data.dataset.fields[data.input_fields[0]].vocab.itos[i.item()] for i in t[0]])
            text_l[i] = new_text
                 
        self.error_tab = pd.DataFrame({'text':text_l,'label':true_l,'preds':preds_l,'loss':loss_l,'score':score_l}).sort_values('loss',ascending=False).reset_index(drop=True)
        self.field = data.dataset.fields[data.input_fields[0]]
        
        self.gym.valid.batch_size = batch_size
        self.gym.valid.batch_size_fn = batch_size_fn
        
    def create_embeddings(self,embeddings=None):
        if embeddings is None:
            emb_present = False
            for m in self.gym.model.modules():
                if isinstance(m,nn.Embedding) and m.num_embeddings==len(self.field.vocab.itos):
                    embeddings = m
                    emb_present = True
            if not emb_present:
                raise ValueError('Cannot find embeddings.')
        self.embeddings = embeddings
        self.error_tab['vec_rep'] = self.error_tab['text'].progress_apply(lambda x: self.embeddings(self.field.process([x]).to(self.gym.device)[0]).sum(dim=0).detach().cpu())
        self.nn_estimator = NearestNeighbors()
        self.nn_estimator.fit(np.stack(self.error_tab['vec_rep'].values))
        
    def find_nn(self,text=None,index=None,n=5,verbose=True):
        if index is not None:
            idx = self.nn_estimator.kneighbors(self.error_tab['vec_rep'][index].reshape(1, -1),n,return_distance=False)
        elif text is not None:
            vector = self.embeddings(self.field.process([text])[0]).sum(dim=0).detach().cpu()
            idx = self.nn_estimator.kneighbors(vector.reshape(1, -1),n,return_distance=False)
        else:
            raise ValueError('Text or index has to be provided.')
        if verbose:
            for index, row in self.error_tab.loc[idx.tolist()[0]].iterrows():
                print(row.score)
                print(row.loss)
                print(row.preds)
                print(row.label)
                print(row.text)
        return self.error_tab[['text','label','preds','loss','score']].loc[idx.tolist()[0]]
            
        
    


from setuptools import setup

setup(
    name='dronnai',
    version='0.1.1',
    description='Neural Networks Training Center by Dronn.com',
    url='https://bitbucket.org/inboxliberatorteam/gym/src/master/',
    author='Dronn AI Team',
    author_email='info@dronn.com',
    # license='MIT',
    packages=['dronnai'],
    install_requires=[
        'numpy',
        'pandas',
        'torch',
        'torchtext',
        'spacy',
        'scikit-learn',
        'tqdm',
        'google.cloud',
        'flatten_dict',
        'sacred'
      ],
      zip_safe=False)